using System.Diagnostics;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using System;

namespace Optim_ModNewton1
{
	public delegate double MyFxDelegate(int nNumVars, ref double[] fX, ref double[] fParam);
	public delegate string SayFxDelegate();
	
	public class COptimModNewton1
	{
		
		MyFxDelegate m_MyFx;
		
		protected double MyFxEx(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fDeltaX, double fLambda)
		{
			int i;
			double[] fXX = new double[nNumVars + 1];
			
			for (i = 0; i < nNumVars; i++)
			{
				fXX[i] = fX[i] + fLambda * fDeltaX[i];
			}

            return m_MyFx(nNumVars, ref fXX, ref fParam);
		}
		
		protected double FirstDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI)
		{
			double fXt, h, Fp, Fm;
			
			fXt = fX[nIdxI];
			h = 0.01 *(1 + Math.Abs(fXt));
			fX[nIdxI] = fXt + h;
			Fp = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt - h;
			Fm = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt;
			return (Fp - Fm) / 2 / h;
		}
		
		protected double SecondDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI, int nIdxJ)
		{
            double fXt, fYt, fHX, fHY, F0, Fp, Fm;
			double Fpp, Fmm, Fpm, Fmp, fResult;
			
			// calculate second derivative?
			if (nIdxI == nIdxJ)
			{
				F0 = m_MyFx(nNumVars, ref fX, ref fParam);
				fXt = fX[nIdxI];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fX[nIdxI] = fXt + fHX;
				Fp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt - fHX;
				Fm = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fResult = (Fp - 2 * F0 + Fm) /Math.Pow(fHX, 2);
			}
			else
			{
				fXt = fX[nIdxI];
				fYt = fX[nIdxJ];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fHY = 0.01 *(1 + Math.Abs(fYt));
				// calculate Fpp
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt + fHY;
				Fpp = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmm
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt - fHY;
				Fmm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fpm
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt - fHY;
				Fpm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmp
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt + fHY;
				Fmp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fX[nIdxJ] = fYt;
				fResult = (Fpp - Fmp - Fpm + Fmm) /(4 * fHX * fHY);
			}
			
			return fResult;
		}
		
		
		protected void GetFirstDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fFirstDerivX)
		{
			
			int i;
			
			for (i = 0; i < nNumVars; i++)
			{
                fFirstDerivX[i] = FirstDeriv(nNumVars, ref fX, ref fParam, i);
			}
		}
		
		protected void GetSecondDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[,] fSecondDerivX)
		{
			
			int i, j;
			
			for (i = 0; i < nNumVars; i++)
			{
				for (j = 0; j < nNumVars; j++)
				{
                    fSecondDerivX[i, j] = SecondDeriv(nNumVars, ref fX, ref fParam, i, j);
				}
			}
		}
		
		protected bool LinSearch_DirectSearch(int nNumVars, ref double[] fX, ref double[] fParam, ref double fLambda, ref double[] fDeltaX, double fInitStep, double fMinStep)
		{
			double F1, F2;

            F1 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda);
			
			do
			{
                F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda + fInitStep);
				if (F2 < F1)
				{
					F1 = F2;
					fLambda += fInitStep;
				}
				else
				{
                    F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda - fInitStep);
					if (F2 < F1)
					{
						F1 = F2;
						fLambda -= fInitStep;
					}
					else
					{
						// reduce search step size
						fInitStep /= 10;
					}
				}
			} while (!(fInitStep < fMinStep));
			
			return true;
			
		}
		
		public double CalcOptim(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fToler, double fEpsFx, int nMaxIter, ref int nIter, ref string sErrorMsg, MyFxDelegate MyFx)
		{
			
			int i;
			double fNorm, fLambda;
			double[] g = new double[nNumVars];
			int[] Index = new int[nNumVars];
			double[] fDeltaX = new double[nNumVars];
			double[,] j = new double[nNumVars, nNumVars];
			bool bStop;
			
			m_MyFx = MyFx;
			
			nIter = 1;
			do
			{
				nIter++;
				if (nIter > nMaxIter)
				{
					sErrorMsg = "Reached maximum iterations limit";
					break;
				}

                GetFirstDerives(nNumVars, ref fX, ref fParam, ref g);
				
				// test if gradient is shallow enough
                fNorm = MatrixLibVb.VectNorm(ref g, 2);
				if (fNorm < fEpsFx)
				{
					break;
				}

                GetSecondDerives(nNumVars, ref fX, ref fParam, ref j);
                MatrixLibVb.MV_LUDecomp(ref j, ref Index, nNumVars);
                MatrixLibVb.MV_LUBackSubst(ref j, ref Index, nNumVars, ref g);
				for (i = 0; i < nNumVars; i++)
				{
					fDeltaX[i] = g[i];
					fX[i] -= fDeltaX[i];
				}
				
				fLambda = 0;
                if (!LinSearch_DirectSearch(nNumVars, ref fX, ref fParam, ref fLambda, ref fDeltaX, 0.1, 0.0001))
				{
					sErrorMsg = "Linear Search failed";
					break;
				}
				
				for (i = 0; i < nNumVars; i++)
				{
					fX[i] += fLambda * fDeltaX[i];
				}
				
				
				bStop = true;
				for (i = 0; i < nNumVars; i++)
				{
					if (Math.Abs(fDeltaX[i]) > fToler[i])
					{
						bStop = false;
						break;
					}
				}
				
			} while (!bStop);
			
			return MyFx(nNumVars, ref fX, ref fParam);
		}
	}
	
}
